/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/

/****************************************************************/
/*								*/
/* Filename:	mutex.c						*/
/* Description:	This file contains the implementation of the 	*/
/*		mutex management functions.			*/
/*								*/
/* Copyright:	Accelerated Encryption Processing Ltd., 2000	*/
/* Author:	Diarmuid O'Neill				*/
/* Product:	AEP1000						*/
/* Date:	12/10/99					*/
/*								*/
/****************************************************************/

#include <pthread.h>

#include "aep_i_defs.h"
#include "aep_i_types.h"

#include "aeptrace.h"
#include "mutex.h"
#include "aepmalloc.h"

/******************************************************************
 *                  Unix Mutex                                    *
 ******************************************************************/
#ifdef AEP_GENERIC

#include <pthread.h>
#include <semaphore.h>

#include <stdlib.h>

AEP_RV AEP_I_CreateMutex(
  AEP_VOID_PTR_PTR ppMutex  /* location to receive ptr to mutex */
)
{
  *ppMutex = AEP_calloc(1,sizeof(pthread_mutex_t));

  if(*ppMutex == NULL_PTR)
    return AEP_R_HOST_MEMORY;
  
  pthread_mutex_init((pthread_mutex_t *)*ppMutex, NULL_PTR);
  
  return AEP_R_OK;
}


AEP_RV AEP_I_DestroyMutex(
  AEP_VOID_PTR pMutex  /* pointer to mutex */
)
{
  pthread_mutex_destroy((pthread_mutex_t *)pMutex);
  AEP_free(pMutex);
  
  return AEP_R_OK;
}

AEP_RV AEP_I_LockMutex(
  AEP_VOID_PTR pMutex  /* pointer to mutex */
)
{
  AEP_RV rv;
  rv = pthread_mutex_lock((pthread_mutex_t *)pMutex);
  if( rv != 0 ) return AEP_R_MUTEX_BAD;
  return AEP_R_OK;
}

AEP_RV AEP_I_UnlockMutex(
  AEP_VOID_PTR pMutex  /* pointer to mutex */
)
{
  AEP_RV rv;
  rv = pthread_mutex_unlock((pthread_mutex_t *)pMutex);
  if( rv != 0 ) return AEP_R_MUTEX_BAD;

  return AEP_R_OK;
}

#elif defined(AEP_Win32)
/********************************************************
 *               Win32 Mutex                            *
 ********************************************************/
#include <windows.h>
#include <stdio.h>
static int depth = 0;

AEP_RV AEP_I_CreateMutex(
  AEP_VOID_PTR_PTR ppMutex  /* location to receive ptr to mutex */
)
{
#if 0
  DWORD err_code;
  AEP_RV rv = AEP_R_OK;
  HANDLE * mutex_field = NULL_PTR;
  
//  if(++depth >0) 
//    {
//      FILE* log;
 //     log = fopen("C:\\mutex.log","a");
 //     fprintf(log,"CreateMutex(Win32): depth:%i",depth);
 //     fclose(log);
 //   }
  
  mutex_field = AEP_malloc(sizeof(int));
  if(!mutex_field)
    {
      rv = AEP_R_HOST_MEMORY;
    	AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "I_CreateMutex", rv, "mutex creation failed");
      return rv;
    }

  *mutex_field = (AEP_VOID_PTR)CreateMutex(NULL_PTR, FALSE, NULL_PTR);

  if(*mutex_field == 0)
    {
      err_code = GetLastError();
      if((err_code == ERROR_NOT_ENOUGH_MEMORY) || ( err_code == ERROR_OUTOFMEMORY ))
	rv = AEP_R_HOST_MEMORY;
      else 
	rv = AEP_R_GENERAL_ERROR;

   	AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "I_CreateMutex", rv, "mutex creation:	GetLastError() :%i", err_code);
    }

  *ppMutex = mutex_field;
  
  return rv;
#else
  *ppMutex = NULL_PTR;
  return AEP_R_OK;
#endif
}


AEP_RV AEP_I_DestroyMutex(
  AEP_VOID_PTR pMutex  /* mutex itself */
)
{
#if 0
  DWORD err_code;

  AEP_RV rv = AEP_R_OK;

//if(--depth >0) 
 // {
 //   FILE* log;
 //   log = fopen("C:\\mutex.log","a");
 //   fprintf(log,"DestroyMutex(Win32): depth:%i",depth);
 //   fclose(log);
 // }

//  CloseHandle(*((HANDLE *)pMutex));
	CloseHandle((HANDLE)pMutex);
  err_code = GetLastError();

  if(err_code != ERROR_SUCCESS)
    switch(err_code)
      {
      case ERROR_INVALID_HANDLE:
	rv = AEP_R_MUTEX_BAD;
	break;
      case ERROR_NOT_ENOUGH_MEMORY:
      case ERROR_OUTOFMEMORY:
	rv = AEP_R_HOST_MEMORY;
	break;
      default:
	rv = AEP_R_GENERAL_ERROR;
      }

  AEP_free(pMutex);

  if(rv != AEP_R_OK)
  {
	AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "I_DestroyMutex", rv, "mutex destruction : GetLastError(): %i", err_code);
	}		   

  return rv;
#else
  return AEP_R_OK;
#endif
}

AEP_RV AEP_I_LockMutex(
  AEP_VOID_PTR pMutex  /* pointer to mutex */
)
{
#if 0
  AEP_RV rv =AEP_R_OK;
  DWORD err_code;

  if(WAIT_FAILED == WaitForSingleObject((HANDLE)pMutex, INFINITE))
    {
      err_code = GetLastError();
      switch(err_code)
	{
	case ERROR_INVALID_HANDLE:
	rv = AEP_R_OK; 
	/*	  rv = AEP_R_MUTEX_BAD; */
	  break;
	case ERROR_NOT_ENOUGH_MEMORY:
	case ERROR_OUTOFMEMORY:
	  rv = AEP_R_HOST_MEMORY;
	  break;
	default:
	  rv = AEP_R_GENERAL_ERROR;
	}
    }

  if(rv != AEP_R_OK)
  {
	AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "I_LockMutex", rv, "mutex locking: GetLastError() : %i", err_code);
	}	   

  return rv;
#else
  return AEP_R_OK;
#endif
}

AEP_RV AEP_I_UnlockMutex(
  AEP_VOID_PTR pMutex  /* pointer to mutex */
)
{
#if 0
    DWORD err_code;

    AEP_RV rv = AEP_R_OK;

  if(!ReleaseMutex((HANDLE)pMutex))
    {
      err_code = GetLastError();
      switch(err_code)
	{
	case ERROR_INVALID_HANDLE:
	rv = AEP_R_OK; 
	/* rv = AEP_R_MUTEX_BAD; */
	  break;
	case ERROR_NOT_ENOUGH_MEMORY:
	case ERROR_OUTOFMEMORY:
	  rv = AEP_R_HOST_MEMORY;
	  break;
	default:
	  rv = AEP_R_GENERAL_ERROR;
	}
    }
  

  if(rv != AEP_R_OK)
  {
	AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_UnlockMutex", rv, "mutex unlocking: GetLastError(): %i", err_code);
	}	   

  return rv;
#else
  return AEP_R_OK;
#endif
}




#else /* ! Win32 */
#error No Mutexes implemented for this architecture!
#endif

#if 0
int SEM_INIT(sem_t *sem, int pshared, unsigned int value)
{
	int err;
	err = sem_init(sem, pshared, value);
	if (err)
	{
		perror("");
		abort();
	}
	return err;
}


int SEM_WAIT(sem_t *arg)	/* Ignore signal interruptions */
{
	while (sem_wait(arg) != 0) 
	{}
	return 0;
}
#endif
