/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/

/****************************************************************/
/*								*/
/* Filename:	primitives.c					*/
/* Description:	This file contains the implementation of the 	*/
/*		primitive cryptographic functions. It includes 	*/
/*		functions to perform modulo exponentiation, 	*/
/*		generate random and prime numbers.	 	*/
/*								*/
/* Copyright:	Accelerated Encryption Processing Ltd., 2000	*/
/* Author:	Diarmuid O'Neill				*/
/* Product:	AEP1000						*/
/* Date:	12/10/99					*/
/*								*/
/****************************************************************/



#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "aep_i_defs.h"
#include "aep_i_types.h"

#include "aep_api.h"
#include "aep_api_globals.h"
#include "aependian.h"
#include "mutex.h"
#include "aepmalloc.h"
#include "connection.h"
#include "SrvInternals.h"
#include "aeptrace.h"
/*	AEP_ModExp:
	This function performs a modular exponentiation on the arguments given

	Inputs	:	Connection Handle  {hConnection}
				Mod Exp Base A  {pA}
				Mod Exp Power P  {pP}
				Mod Exp Modulus  {pN}
				Pointer to a Transaction handle (either a valid ptr or NULL)  {pidTransId}

	Outputs	:	Result of the modular exp.  {pResult}
				Transaction handle is returned if a valid pointer was passed in pidTransId
*/
AEP_SLIB_CALLING_CONVENTION AEP_RV AEP_ModExp(
        AEP_CONNECTION_HNDL		hConnection,
		AEP_VOID_PTR			pA,
		AEP_VOID_PTR			pP,
		AEP_VOID_PTR			pN,
		AEP_VOID_PTR			pResult,
		AEP_TRANSACTION_ID*		pidTransId
      )
{
	AEP_BYTEBLOCK	commandblock, replyblock;
	AEP_U32			datalen, index, sizeA, sizeP, sizeN;
	AEP_MSG_HDR_PTR	message_hdr;
	AEP_U32			tran_id;
 
	AEP_RV			rv = AEP_R_OK;

	AEP_TRACE(API, TRACE_DEBUG, FN_IN, "AEP_ModExp", AEP_R_OK, "Entered...");

/* Protect the global flags variable */ 
	_LOCK(&AEPAPI_flags_mutex);

  /* make sure we are initialized */
	if (!(AEP_I_global_flags & AEP_IGF_INITIALIZED))
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_ModExp", AEP_R_AEPAPI_NOT_INITIALIZED, "Error: API Already Initialised");
		_UNLOCK(&AEPAPI_flags_mutex);
		return AEP_R_AEPAPI_NOT_INITIALIZED;
	}

  /* Get the Tran ID to assign to this request*/
	rv = AEP_I_NewTranID(&tran_id);
	if(rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_ModExp", rv, "AEP_I_ModExp failed on connection %l", hConnection);
			
		_UNLOCK(&AEPAPI_flags_mutex);
		return rv;
	}

/*Finished with the flags */
	_UNLOCK(&AEPAPI_flags_mutex);

  /*Get the size in bytes of each of the big nums required to do the mod exp*/
	rv = AEP_I_ext_functions.AEP_I_GetBigNumSize(pA, &sizeA);
	if(rv != AEP_R_OK)
		return rv;

	rv = AEP_I_ext_functions.AEP_I_GetBigNumSize(pP, &sizeP);
	if(rv != AEP_R_OK)
		return rv;

	rv = AEP_I_ext_functions.AEP_I_GetBigNumSize(pN, &sizeN);
	if(rv != AEP_R_OK)
		return rv;

  /*Calculate the size of the data*/
  	datalen = sizeof(AEP_U32)*3 + sizeA + sizeP + sizeN;

  /*How big is this byteblock going to be?*/
	commandblock.Len = HDR_SIZE + datalen;

  /*Allocate a buffer big enough to hold the message (Header + Data)*/
	commandblock.ptr=(AEP_U8_PTR)AEP_malloc(commandblock.Len);
	if(commandblock.ptr == NULL_PTR)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_ModExp", AEP_R_HOST_MEMORY, "Error: Host Memory Allocation Failure");
		return AEP_R_HOST_MEMORY;
	}
 
  /*We make this assignemt so we can refer to our message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)commandblock.ptr;

  /*Copy the command identifier, the data length and the status*/
	message_hdr->command = AEP_C_MODEXP;
	
	message_hdr->status = AEP_S_INIT;

  /*Copy the transaction handle for this request*/
	message_hdr->trans_hndl = (AEP_U64)getpid() << 32 | tran_id;

  /*This command is for the accelerator device*/
	message_hdr->flags = AEP_F_NOCTXCMD | AEP_F_ASIC;

  /*Set the message length field*/
	message_hdr->datalen = datalen;

  /*Copy the command's data into the message*/
	index = HDR_DATA_OFFSET;

	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)(sizeA);
	BEndToLEnd(commandblock.ptr + index, 1);
	index+=sizeof(AEP_U32);

	rv = AEP_I_ext_functions.AEP_I_MakeAEPBigNum(pA, sizeA, commandblock.ptr + index);
	if(rv != AEP_R_OK)
		return rv;

	index += sizeA;

	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)(sizeP);
	BEndToLEnd(commandblock.ptr + index, 1);
	index+=sizeof(AEP_U32);	
	
	rv = AEP_I_ext_functions.AEP_I_MakeAEPBigNum(pP, sizeP, commandblock.ptr + index);
	if(rv != AEP_R_OK)
		return rv;

	index += sizeP;

	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)(sizeN);
	BEndToLEnd(commandblock.ptr + index, 1);
	index+=sizeof(AEP_U32);

	rv = AEP_I_ext_functions.AEP_I_MakeAEPBigNum(pN, sizeN, commandblock.ptr + index);
	if(rv != AEP_R_OK)
		return rv;

  /*Is this transaction to be synch or async?*/
	if (pidTransId==NULL)
	{
		rv = AEP_I_Transact(hConnection, &commandblock, &replyblock);

		if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_ModExp", rv, "Error: AEP_I_ModExp failed on connection %l", hConnection);

			/*Free the data block that was used to hold the message*/
			AEP_free(commandblock.ptr);
			return rv;
		}

	  /*Copy the returned result into the required result argument*/
		LEndToBEnd(replyblock.ptr + HDR_DATA_OFFSET, 1);

		rv = AEP_I_ext_functions.AEP_I_ConvertAEPBigNum(pResult, 
					*(AEP_U32*)(replyblock.ptr + HDR_DATA_OFFSET),
					replyblock.ptr+HDR_DATA_OFFSET+4);
		if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_ModExp", rv, "Error: AEP_I_ModExp failed on connection %l", hConnection);

			/*Free the data block that was used to hold the message*/
			AEP_free(commandblock.ptr);

			return rv;
		}
					
	  /*Free the data block that was used to hold the reply*/
		AEP_free(replyblock.ptr);
	}
	else
	{
		rv = AEP_I_Submit(hConnection, &commandblock);

		*pidTransId = tran_id;

		if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_ModExp", rv, "Error: AEP_I_ModExp failed on connection %l", hConnection);

		  /*Free the data block that was used to hold the message*/
			AEP_free(commandblock.ptr);
			return rv;
		}
	}
  /*Free the data block that was used to hold the message*/
	AEP_free(commandblock.ptr);
	
	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "AEP_ModExp", AEP_R_OK, "... Leaving");

	return rv; 
}

/*	AEP_ModExpCrt:
	This function performs a modular exponentiation, and may use CRT
	to make the calculation quicker.

	Inputs	:	Connection Handle  {hConnection}
				Mod Exp Base A  {pA}
				Mod Exp First Modulus Factor P	{pP}
				Mod Exp Second Modulus Factor Q	{pQ}
				Mod Exp D Mod P-1   {pDmp1}
				Mod Exp D Mod Q-1	{pDmq1}
				Mod Exp Q**-1 Mod P	{pIqmp}
				Pointer to a Transaction handle (either a valid ptr or NULL)  {pidTransId}

	Outputs	:	Result of the modular exp.  {pResult}
				Transaction handle is returned if a valid pointer was passed in pidTransId
*/
AEP_SLIB_CALLING_CONVENTION AEP_RV AEP_ModExpCrt(
        AEP_CONNECTION_HNDL		hConnection,
		AEP_VOID_PTR			pA,
		AEP_VOID_PTR			pP,
		AEP_VOID_PTR			pQ,
		AEP_VOID_PTR			pDmp1,
		AEP_VOID_PTR			pDmq1,
		AEP_VOID_PTR			pIqmp,
		AEP_VOID_PTR			pResult,
		AEP_TRANSACTION_ID*		pidTransId
      )
{
	AEP_BYTEBLOCK		commandblock, replyblock;
	AEP_U32			datalen, index, sizeA, sizeP, sizeQ;
	AEP_U32			sizeDmp1, sizeDmq1, sizeIqmp;
	AEP_MSG_HDR_PTR	message_hdr;
	AEP_U32			tran_id;
 
	AEP_RV			rv = AEP_R_OK;

	AEP_TRACE(API, TRACE_DEBUG, FN_IN, "AEP_ModExpCrt", AEP_R_OK, "Entered.... ");

/* Protect the global flags variable */ 
	_LOCK(&AEPAPI_flags_mutex);

  /* make sure we are initialized */
	if (!(AEP_I_global_flags & AEP_IGF_INITIALIZED))
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_ModExpCrt", AEP_R_AEPAPI_NOT_INITIALIZED, "Error: API Already Initialised ");
		_UNLOCK(&AEPAPI_flags_mutex);
		return AEP_R_AEPAPI_NOT_INITIALIZED;
	}

  /* Get the Tran ID to assign to this request*/
	rv = AEP_I_NewTranID(&tran_id);
	if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_ModExpCrt", rv, 	"AEP_I_ModExpCrt failed on connection %l", hConnection);
			_UNLOCK(&AEPAPI_flags_mutex);
			return rv;
		}

/*Finished with the flags */
	_UNLOCK(&AEPAPI_flags_mutex);

  /*Get the size in bytes of each of the big nums required to do the mod exp*/
	rv = AEP_I_ext_functions.AEP_I_GetBigNumSize(pA, &sizeA);
	if(rv != AEP_R_OK)
		return rv;

	rv = AEP_I_ext_functions.AEP_I_GetBigNumSize(pP, &sizeP);
	if(rv != AEP_R_OK)
		return rv;

	rv = AEP_I_ext_functions.AEP_I_GetBigNumSize(pQ, &sizeQ);
	if(rv != AEP_R_OK)
		return rv;

	rv = AEP_I_ext_functions.AEP_I_GetBigNumSize(pDmp1, &sizeDmp1);
	if(rv != AEP_R_OK)
		return rv;

	rv = AEP_I_ext_functions.AEP_I_GetBigNumSize(pDmq1, &sizeDmq1);
	if(rv != AEP_R_OK)
		return rv;

	rv = AEP_I_ext_functions.AEP_I_GetBigNumSize(pIqmp, &sizeIqmp);
	if(rv != AEP_R_OK)
		return rv;

  /*Calculate the size of the data*/
  	datalen = sizeof(AEP_U32)*9 + sizeA + sizeP + sizeQ + sizeDmp1 + sizeDmq1 + sizeIqmp;

  /*How big is this byteblock going to be?*/
	commandblock.Len = HDR_SIZE + datalen;

  /*Allocate a buffer big enough to hold the message (Header + Data)*/
	commandblock.ptr=(AEP_U8_PTR)AEP_malloc(commandblock.Len);
	if(commandblock.ptr == NULL_PTR)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_ModExpCrt", AEP_R_HOST_MEMORY, "Error: Host Memory Allocation Failure");
		return AEP_R_HOST_MEMORY;
	}
 
  /*We make this assignemt so we can refer to our message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)commandblock.ptr;

  /*Copy the command identifier, the data length and the status*/
	message_hdr->command = AEP_C_MODEXPCRT;
	
	message_hdr->status = AEP_S_INIT;

  /*Copy the transaction handle for this request*/
	message_hdr->trans_hndl = (AEP_U64)getpid() << 32 | tran_id;

  /*This command is for the accelerator device*/
	message_hdr->flags = AEP_F_NOCTXCMD | AEP_F_ASIC;

  /*Set the message length field*/
	message_hdr->datalen = datalen;

  /*Copy the command's data into the message*/
	index = HDR_DATA_OFFSET;

  /*Set up the number of factors*/
	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)2;
	BEndToLEnd(commandblock.ptr + index, 1);	
	index+=sizeof(AEP_U32);

  /*Leave space for 2 pointer values to be used on the board*/
	index+=sizeof(AEP_U32) + sizeof(AEP_U32);

  /*Fill in the various big nums*/
	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)(sizeA);
	BEndToLEnd(commandblock.ptr + index, 1);
	index+=sizeof(AEP_U32);

	rv = AEP_I_ext_functions.AEP_I_MakeAEPBigNum(pA, sizeA, commandblock.ptr + index);
	if(rv != AEP_R_OK)
		return rv;

	index += sizeA;

	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)(sizeP);
	BEndToLEnd(commandblock.ptr + index, 1);
	index+=sizeof(AEP_U32);	
	
	rv = AEP_I_ext_functions.AEP_I_MakeAEPBigNum(pP, sizeP, commandblock.ptr + index);
	if(rv != AEP_R_OK)
		return rv;

	index += sizeP;

	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)(sizeQ);
	BEndToLEnd(commandblock.ptr + index, 1);
	index+=sizeof(AEP_U32);

	rv = AEP_I_ext_functions.AEP_I_MakeAEPBigNum(pQ, sizeQ, commandblock.ptr + index);
	if(rv != AEP_R_OK)
		return rv;

	index += sizeQ;

	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)(sizeDmp1);
	BEndToLEnd(commandblock.ptr + index, 1);
	index+=sizeof(AEP_U32);

	rv = AEP_I_ext_functions.AEP_I_MakeAEPBigNum(pDmp1, sizeDmp1, commandblock.ptr + index);
	if(rv != AEP_R_OK)
		return rv;

	index += sizeDmp1;

	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)(sizeDmq1);
	BEndToLEnd(commandblock.ptr + index, 1);
	index+=sizeof(AEP_U32);	
	
	rv = AEP_I_ext_functions.AEP_I_MakeAEPBigNum(pDmq1, sizeDmq1, commandblock.ptr + index);
	if(rv != AEP_R_OK)
		return rv;

	index += sizeDmq1;

	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)(sizeIqmp);
	BEndToLEnd(commandblock.ptr + index, 1);
	index+=sizeof(AEP_U32);

	rv = AEP_I_ext_functions.AEP_I_MakeAEPBigNum(pIqmp, sizeIqmp, commandblock.ptr + index);
	if(rv != AEP_R_OK)
		return rv;

  /*Is this transaction to be synch or async?*/
	if (pidTransId==NULL)
	{
		rv = AEP_I_Transact(hConnection, &commandblock, &replyblock);

		if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_ModExpCrt", rv, "Error: AEP_I_ModExpCRT failed on connection %l", hConnection);

			/*Free the data block that was used to hold the message*/
			AEP_free(commandblock.ptr);
			return rv;
		}

	  /*Copy the returned result into the required result argument*/
		LEndToBEnd(replyblock.ptr + HDR_DATA_OFFSET, 1);
		
		rv = AEP_I_ext_functions.AEP_I_ConvertAEPBigNum(pResult, 
					*(AEP_U32*)(replyblock.ptr + HDR_DATA_OFFSET),
					replyblock.ptr+HDR_DATA_OFFSET+4);

		if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_ModExpCrt", rv, "Error: AEP_I_ModExpCrt failed on connection %l", hConnection);		

			/*Free the data block that was used to hold the message*/
			AEP_free(commandblock.ptr);
			return rv;
		}
					
	  /*Free the data block that was used to hold the reply*/
		AEP_free(replyblock.ptr);
	}
	else
	{
		rv = AEP_I_Submit(hConnection, &commandblock);

		*pidTransId = tran_id;

		if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_ModExpCrt", rv, "Error: AEP_I_ModExpCrt failed on connection %l ", hConnection);		

		  /*Free the data block that was used to hold the message*/
			AEP_free(commandblock.ptr);
			return rv;
		}
	}
  /*Free the data block that was used to hold the message*/
	AEP_free(commandblock.ptr);
	
	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "AEP_ModExpCrt", AEP_R_OK, "... Leaving");

	return rv; 
}

 
/*	AEP_GenPrime:
	This function generates a prime number of given length

	Inputs	:	Connection Handle  {hConnection}
				Length (In bits)  {Len}
				Pointer to a Transaction handle (either a valid ptr or NULL)  {pidTransId}

	Outputs	:	The prime number.  {pResult}
				Transaction handle is returned if a valid pointer was passed in pidTransId
*/
AEP_SLIB_CALLING_CONVENTION AEP_RV AEP_GenPrime(
        AEP_CONNECTION_HNDL		hConnection,
		AEP_U32					Len,
		AEP_VOID_PTR			pResult,
		AEP_TRANSACTION_ID*		pidTransId
      )
{
	AEP_BYTEBLOCK	commandblock, replyblock;
	AEP_U32		datalen;
	AEP_MSG_HDR_PTR	message_hdr;
	AEP_U32		tran_id;

 	AEP_RV			rv = AEP_R_OK;
  
	AEP_TRACE(API, TRACE_DEBUG, FN_IN, "AEP_GenPrime", AEP_R_OK, "Entered...");
	
/* Protect the global flags variable */ 
	_LOCK(&AEPAPI_flags_mutex);

  /* make sure we are initialized */
	if (!(AEP_I_global_flags & AEP_IGF_INITIALIZED))
	{
		_UNLOCK(&AEPAPI_flags_mutex);
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GenPrime", AEP_R_AEPAPI_NOT_INITIALIZED, "Error: API not initialized");

		return AEP_R_AEPAPI_NOT_INITIALIZED;
	}

  /*Check for a zero length request*/
	if(Len == 0)
	{
		_UNLOCK(&AEPAPI_flags_mutex);
		return rv;
	}

  /* Get the Tran ID to assign to this request*/
	rv = AEP_I_NewTranID(&tran_id);
	if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GenPrime", rv, "Error: AEP_GenPrime failed on connection %l", hConnection);
			_UNLOCK(&AEPAPI_flags_mutex);
			return rv;
		}

/*Finished with the flags */
	_UNLOCK(&AEPAPI_flags_mutex);
  
  	datalen = sizeof(AEP_U32) + Len/8;

  /*Make sure the len is a multiple of bytes*/
  /*No need if we're checking for multiple of words anyway*/
/*	if (Len % 8 > 0)
		datalen++;*/

  /*Make sure the len is a multiple of words*/
	if (datalen % sizeof(AEP_U32) !=0)
		datalen += sizeof(AEP_U32) - (datalen % sizeof(AEP_U32));

  /*How big is this byteblock going to be?*/
	commandblock.Len = HDR_SIZE + datalen;

  /*Allocate a buffer big enough to hold the message (Header + Data)*/
	commandblock.ptr=(AEP_U8_PTR)AEP_malloc(commandblock.Len);
	if(commandblock.ptr == NULL_PTR)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GenPrime", AEP_R_HOST_MEMORY, "Error: Host memory Allocation Failure");
		return AEP_R_HOST_MEMORY;
	}

 /*We make this assignemt so we can refer to our message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)commandblock.ptr;

  /*Copy the command identifier, the data length and the status*/
	message_hdr->command = AEP_C_GENPRIME;
	
	message_hdr->status = AEP_S_INIT;

  /*Copy the transaction handle for this request*/
	message_hdr->trans_hndl = (AEP_U64)getpid() << 32 | tran_id;

  /*This command is for the accelerator device*/
	message_hdr->flags = AEP_F_NOCTXCMD | AEP_F_SA;

	message_hdr->datalen = datalen;

  /*Copy the command's data */
	memcpy(commandblock.ptr + HDR_DATA_OFFSET, &Len, 4);

  /*Adjust the endianess of the data*/
	BEndToLEnd(commandblock.ptr + HDR_DATA_OFFSET, 1);

  /*Is this transaction to be synch or async?*/
	if (pidTransId==NULL)
	{
		rv = AEP_I_Transact(hConnection, &commandblock, &replyblock);
		if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GenPrime", rv, "Error: AEP_GenPrime failed on connection %l", hConnection);
			
			/*Free the data block that was used to hold the message*/
			AEP_free(commandblock.ptr);
			return rv;
		}
	  
	  /*Copy the returned result into the required result argument*/
		BEndToLEnd(replyblock.ptr + HDR_DATA_OFFSET, 1);

		rv = AEP_I_ext_functions.AEP_I_ConvertAEPBigNum(pResult, 
					*(AEP_U32*)(replyblock.ptr + HDR_DATA_OFFSET),
					replyblock.ptr+HDR_DATA_OFFSET+4);
		if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GenPrime", rv, "Error: AEP_GenPrime failed on connection %l", hConnection);

			/*Free the data block that was used to hold the message*/
			AEP_free(commandblock.ptr);
			return rv;
		}
	  
	  /*Free the data block that was used to hold the reply*/
		AEP_free(replyblock.ptr);
	}
	else
	{
		rv = AEP_I_Submit(hConnection, &commandblock);

		*pidTransId = tran_id;

		if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GenPrime", rv, "Error: AEP_GenPrime failed on connection %l", hConnection);

		  /*Free the data block that was used to hold the message*/
			AEP_free(commandblock.ptr);
			return rv;
		}
	}

  /*Free the data block that was used to hold the message*/
	AEP_free(commandblock.ptr);

	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "AEP_GenPrime", AEP_R_OK, "...Leaving");

	return rv; 
}

/*	AEP_GenRandom:
	This function generates a random number of given length

	Inputs	:	Connection Handle  {hConnection}
				Length (In bytes)  {Len}
				Type of Random number required {Type}
				Pointer to a Transaction handle (either a valid ptr or NULL)  {pidTransId}

	Outputs	:	The random number.  {pResult}
				Transaction handle is returned if a valid pointer was passed in pidTransId
*/
AEP_SLIB_CALLING_CONVENTION AEP_RV AEP_GenRandom(
        	AEP_CONNECTION_HNDL		hConnection,
		AEP_U32				Len,
		AEP_U32				Type,
		AEP_VOID_PTR			pResult,
		AEP_TRANSACTION_ID*		pidTransId
      )
{
	AEP_BYTEBLOCK	commandblock, replyblock;
	AEP_U32		datalen;
	AEP_MSG_HDR_PTR	message_hdr;
	AEP_U32		tran_id, index;

 
	AEP_RV			rv = AEP_R_OK;
  
	AEP_TRACE(API, TRACE_DEBUG, FN_IN, "AEP_GenRand", AEP_R_OK, "Entered...");

/* Protect the global flags variable */ 
	_LOCK(&AEPAPI_flags_mutex);

  /* make sure we are initialized */
	if (!(AEP_I_global_flags & AEP_IGF_INITIALIZED))
	{
		_UNLOCK(&AEPAPI_flags_mutex);
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GenRand", AEP_R_AEPAPI_NOT_INITIALIZED, "Error : API already initialized");		
		
		return AEP_R_AEPAPI_NOT_INITIALIZED;
	}

  /*Check for a zero length request*/
	if(Len == 0)
	{
		_UNLOCK(&AEPAPI_flags_mutex);
		return rv;
	}

  /* Get the Tran ID to assign to this request*/
	rv = AEP_I_NewTranID(&tran_id);
	if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GenRand", rv, "Error : failed on connection %l", hConnection);				
			
			_UNLOCK(&AEPAPI_flags_mutex);
			return rv;
		}

/*Finished with the flags */
	_UNLOCK(&AEPAPI_flags_mutex);
  
  /*Calculate the size of the data*/
  	datalen = sizeof(AEP_U32) + sizeof(AEP_U32) + Len;

  /*How big is this byteblock going to be?  Include len of required no.*/
	commandblock.Len = HDR_SIZE + datalen;

  /*Allocate a buffer big enough to hold the message (Header + Data)*/
	commandblock.ptr=(AEP_U8_PTR)AEP_malloc(commandblock.Len);
	if(commandblock.ptr == NULL_PTR)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GenRand", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure");				
		return AEP_R_HOST_MEMORY;
	}

 /*We make this assignemt so we can refer to our message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)commandblock.ptr;

  /*Copy the command identifier, the data length and the status*/
	message_hdr->command = AEP_C_GENRANDOM;
	
	message_hdr->status = AEP_S_INIT;

  /*Copy the transaction handle for this request*/
	message_hdr->trans_hndl = (AEP_U64)getpid() << 32 | tran_id;

	message_hdr->flags = AEP_F_NOCTXCMD | AEP_F_SA;

	message_hdr->datalen = datalen;

  /*Copy the command's data into the message*/
	index = HDR_DATA_OFFSET;

  /*Copy the command's data */
	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)(Len);
	
  /*Adjust the endianess of the data*/
	BEndToLEnd(commandblock.ptr + index, 1);
	
	index+=sizeof(AEP_U32);
	
  /*Copy the command's data */
	*((AEP_U32_PTR)(commandblock.ptr + index)) = (AEP_U32)(Type);
	
  /*Adjust the endianess of the data*/
	BEndToLEnd(commandblock.ptr + index, 1);

  /*Is this transaction to be synch or async?*/
	if (pidTransId==NULL)
	{
		rv = AEP_I_Transact(hConnection, &commandblock, &replyblock);

		if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GenRand", rv, "Error : failed on connection %l", hConnection);					
			
			/*Free the data block that was used to hold the message*/
			AEP_free(commandblock.ptr);
  
			return rv;
		}
	  
	  /*Copy the returned result into the required result argument*/
		BEndToLEnd(replyblock.ptr + HDR_DATA_OFFSET, 1);

	 	memcpy(pResult, replyblock.ptr + HDR_DATA_OFFSET + 4,
			*(AEP_U32*)(replyblock.ptr + HDR_DATA_OFFSET));

	  /*Free the data block that was used to hold the reply*/
		AEP_free(replyblock.ptr);

	}
	else
	{
		rv = AEP_I_Submit(hConnection, &commandblock);

		*pidTransId = tran_id;

		if(rv != AEP_R_OK)
		{
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GenRand", rv, "Error : failed on connection %l", hConnection);				

		  /*Free the data block that was used to hold the message*/
			AEP_free(commandblock.ptr);
			return rv;
		}
	}

  /*Free the data block that was used to hold the message*/
	AEP_free(commandblock.ptr);

	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "AEP_GenRand", AEP_R_OK, "....Leaving");				

	return rv; 
}
  
/*	AEP_QueryReply:
	This function checks to see if the results of a previous command a ready.  IF they are,
	it returns them.

	Inputs	:	Connection Handle  {hConnection}
				Transaction handle  {idTransID}

	Outputs	:	The result pointed to by pResult
*/   
AEP_SLIB_CALLING_CONVENTION AEP_RV AEP_QueryReply(
        AEP_CONNECTION_HNDL		hConnection,
		AEP_TRANSACTION_ID		idTransID,
		AEP_VOID_PTR			pResult
      )
{
	AEP_BYTEBLOCK	commandblock, replyblock;
 
	AEP_RV			rv = AEP_R_OK;

	AEP_MSG_HDR_PTR	message_hdr;
	AEP_MSG_HDR_PTR response_hdr;
  
	AEP_TRACE(DAEMON, TRACE_DEBUG, FN_IN, "AEP_QueryReply", AEP_R_OK, "Entered...");

/* Protect the global flags variable */ 
	_LOCK(&AEPAPI_flags_mutex);

  /* make sure we are initialized */
	if (!(AEP_I_global_flags & AEP_IGF_INITIALIZED))
	{
		_UNLOCK(&AEPAPI_flags_mutex);
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "QueryReply", AEP_R_AEPAPI_NOT_INITIALIZED, "Error : API already initialized");
		return AEP_R_AEPAPI_NOT_INITIALIZED;
	}

/*Finished with the flags */
	_UNLOCK(&AEPAPI_flags_mutex);

  /*How big is this byteblock going to be?*/
	commandblock.Len = HDR_SIZE;

  /*Allocate a buffer big enough to hold the message (Header + Data)*/
	commandblock.ptr=(AEP_U8_PTR)AEP_calloc(1,commandblock.Len);
	if(commandblock.ptr == NULL_PTR)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "QueryReply", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure");
		return AEP_R_HOST_MEMORY;
	}

  /*We make this assignemt so we can refer to our message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)commandblock.ptr;

  /*Copy the command identifier, the data length and the status*/
	message_hdr->command = AEP_C_QUERY;
	message_hdr->flags = AEP_F_SRVCMD;
	message_hdr->trans_hndl = (AEP_U64)getpid() << 32 | idTransID;

	rv = AEP_I_Query(hConnection, &commandblock, &replyblock);

  /*Reply is not yet available*/
	if (rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_DEBUG, FN_NONE, "QueryReply", rv, "Reply not yet available");

		return rv;
	}

	response_hdr = (AEP_MSG_HDR_PTR)replyblock.ptr;

  /*This happens to be the same for modexp genran and gen prime.  Because they all return
  big numbers.  This will be reply specific => as much functions are written this funciton
  will get bigger*/

  /*The command this reply was generated for is stored in the message header.  The following messy
	expression simply equates to "the value at the address replyblock.ptr plus the command offset" 
	(which we want to view as an AEP_U32)*/

	switch (response_hdr->command)
	{
		case AEP_C_MODEXP:
		case AEP_C_GENPRIME:
		case AEP_C_MODEXPCRT:
		  /*Copy the returned result into the required result argument*/
		  	BEndToLEnd(replyblock.ptr + HDR_DATA_OFFSET, 1);

			rv = AEP_I_ext_functions.AEP_I_ConvertAEPBigNum(pResult, 
				*(AEP_U32*)(replyblock.ptr + HDR_DATA_OFFSET),
				replyblock.ptr+HDR_DATA_OFFSET+4);
	
			if(rv != AEP_R_OK)
			{
				AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "QueryReply", rv, "Error : failed on connection %l", hConnection);

				/*Free the data block that was used to hold the message*/
				AEP_free(commandblock.ptr);
				return rv;
			}
			break;

		case AEP_C_GENRANDOM:
		  /*Copy the returned result into the required result argument*/
			BEndToLEnd(replyblock.ptr + HDR_DATA_OFFSET, 1);

		 	memcpy(pResult, replyblock.ptr + HDR_DATA_OFFSET + 4,
				*(AEP_U32*)(replyblock.ptr + HDR_DATA_OFFSET));
			break;
	}

  /*Free the data block that was used to hold the message*/
	AEP_free(commandblock.ptr);

  /*Free the data block that was used to hold the reply*/
	AEP_free(replyblock.ptr);

	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "QueryReply", AEP_R_OK, "...Leaving");

	return rv; 
} 	


/*	AEP_AwaitReply:
	This function checks to see if the results of a previous command a ready.  IF they are,
	it returns them.

	Inputs	:	Connection Handle  {hConnection}
				Transaction handle  {idTransID}

	Outputs	:	The result pointed to by pResult
*/   
AEP_SLIB_CALLING_CONVENTION AEP_RV AEP_AwaitReply(
        AEP_CONNECTION_HNDL		hConnection,
		AEP_TRANSACTION_ID		idTransID,
		AEP_VOID_PTR			pResult
      )
{
	AEP_BYTEBLOCK	commandblock, replyblock;
 
	AEP_RV			rv = AEP_R_OK;

	AEP_MSG_HDR_PTR	message_hdr;
	AEP_MSG_HDR_PTR response_hdr;


	AEP_TRACE(DAEMON, TRACE_DEBUG, FN_IN, "AEP_AwaitReply", AEP_R_OK, "Entered...");

/* Protect the global flags variable */ 
	_LOCK(&AEPAPI_flags_mutex);

  /* make sure we are initialized */
	if (!(AEP_I_global_flags & AEP_IGF_INITIALIZED))
	{
		_UNLOCK(&AEPAPI_flags_mutex);
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_AwaitReply", AEP_R_AEPAPI_NOT_INITIALIZED, "Error : API Already Initialized");
		return AEP_R_AEPAPI_NOT_INITIALIZED;
	}

/*Finished with the flags */
	_UNLOCK(&AEPAPI_flags_mutex);

  /*How big is this byteblock going to be?*/
	commandblock.Len = HDR_SIZE;

  /*Allocate a buffer big enough to hold the message (Header + Data)*/
	commandblock.ptr=(AEP_U8_PTR)AEP_calloc(1,commandblock.Len);
	if(commandblock.ptr == NULL_PTR)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_AwaitReply", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure");
		return AEP_R_HOST_MEMORY;
	}

 /*We make this assignemt so we can refer to our message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)commandblock.ptr;

  /*Copy the command identifier, the data length and the status*/
	message_hdr->command = AEP_C_WAIT;
	message_hdr->flags = AEP_F_SRVCMD;
	message_hdr->trans_hndl = (AEP_U64)getpid() << 32 | idTransID;

	rv = AEP_I_Wait(hConnection, &commandblock, &replyblock);

  /*An error occured while waiting for the reply*/
	if (rv != AEP_R_OK)
	{
	  /*Free the data block that was used to hold the message*/
		AEP_free(commandblock.ptr);

		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_AwaitReply", rv, "Error waiting for the Reply");
			
		return rv;	
	}

	response_hdr = (AEP_MSG_HDR_PTR)replyblock.ptr;

  /*This happens to be the same for modexp genran and gen prime.  Because they all return
  big numbers.  This will be reply specific => as much functions are written this funciton
  will get bigger*/

  /*The command this reply was generated for is stored in the message header.  The following messy
	expression simply equates to "the value at the address replyblock.ptr plus the command offset" 
	(which we want to view as an AEP_U32)*/

	switch (response_hdr->command)
	{
		case AEP_C_MODEXP:
		case AEP_C_GENPRIME:
		case AEP_C_MODEXPCRT:
		  /*Copy the returned result into the required result argument*/
		  	BEndToLEnd(replyblock.ptr + HDR_DATA_OFFSET, 1);

			rv = AEP_I_ext_functions.AEP_I_ConvertAEPBigNum(pResult, 
				*(AEP_U32*)(replyblock.ptr + HDR_DATA_OFFSET),
				replyblock.ptr+HDR_DATA_OFFSET+4);
		
			if(rv != AEP_R_OK)
			{
				AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_AwaitReply", rv, "Error : failed on connection %l",hConnection);


				/*Free the data block that was used to hold the message*/
				AEP_free(commandblock.ptr);
				return rv;
			}
	
			break;

		case AEP_C_GENRANDOM:
		  /*Copy the returned result into the required result argument*/
			BEndToLEnd(replyblock.ptr + HDR_DATA_OFFSET, 1);

		 	memcpy(pResult, replyblock.ptr + HDR_DATA_OFFSET + 4,
				*(AEP_U32*)(replyblock.ptr + HDR_DATA_OFFSET));
		break;

	}

	/*Free the data block that was used to hold the message*/
	AEP_free(commandblock.ptr);

  /*Free the data block that was used to hold the reply*/
	AEP_free(replyblock.ptr);
	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "AEP_AwaitReply", AEP_R_OK, "...Leaving");

	return rv; 
} 	
 

 
