/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "aep_types.h"
#include "aep_defs.h"
#include "aep_api.h"
#include "aep_lib.h"
#include "aep_i_types.h"
#include "aep_i_defs.h"
#include "aep_api_globals.h"
#include "aep_version.h"
#include "aepmalloc.h"
#include "mutex.h"
#include "aependian.h"
#include "trace.h"
#include "api_version.h"
#include "aeptrace.h"
#include "SrvInternals.h"
#include "connection.h"


/*	AEP_GetVersion:
	This function gets information about the various software versions on the system

	Inputs	:	Connection handle {hConnection}
	
	Outputs	:	Pointer to struct which contains the version numbers {pVersions}
*/
AEP_SLIB_CALLING_CONVENTION AEP_RV AEP_GetVersion(AEP_CONNECTION_HNDL	hConnection,
							AEP_VERSION *pVersions )
{
	
	AEP_RV			rv = AEP_R_OK;
	AEP_U32			datalen;
	AEP_MSG_HDR_PTR		message_hdr;
	AEP_U32			tran_id;
	AEP_BYTEBLOCK		commandblock, 
				replyblock;
	AEP_U8_PTR		replyIndex;
	
	
	AEP_TRACE(API, TRACE_DEBUG, FN_IN, "AEP_GetVersion", AEP_R_OK, "Entered...");				
	
	/* Protect the global flags variable */ 
	_LOCK(&AEPAPI_flags_mutex);
	
	/* make sure we are initialized */
	if (!(AEP_I_global_flags & AEP_IGF_INITIALIZED))
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GetVersion", AEP_R_AEPAPI_NOT_INITIALIZED, "Error: API not Initialised");				
		_UNLOCK(&AEPAPI_flags_mutex);
		return AEP_R_AEPAPI_NOT_INITIALIZED;
	}
	
	/*Step 1. Send a message to the target to get the version of the target software*/
	rv = AEP_I_NewTranID(&tran_id);
	if(rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GetVersion", rv, "AEP_GetVersion failed on connection %l", hConnection);
		_UNLOCK(&AEPAPI_flags_mutex);
		return rv;
	}
	
	/*Finished with the flags */
	_UNLOCK(&AEPAPI_flags_mutex);

	/*Leave room for the result.*/
	datalen=sizeof(AEP_TARGET_VERSION);

	/*How big is this byteblock going to be?*/
	commandblock.Len = HDR_SIZE + datalen;

  /*Allocate a buffer big enough to hold the message (Header + Data)*/
	commandblock.ptr=(AEP_U8_PTR)AEP_malloc(commandblock.Len);
	if(commandblock.ptr == NULL_PTR)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GetVersion", AEP_R_HOST_MEMORY, "Error: Host Memory Allocation Failure");				
		return AEP_R_HOST_MEMORY;
	}
 
  /*We make this assignemt so we can refer to our message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)commandblock.ptr;

  /*Copy the command identifier, the data length and the status*/
	message_hdr->command = AEP_C_GET_VERSION;
	
	message_hdr->status = AEP_S_INIT;

  /*Copy the transaction handle for this request*/
	message_hdr->trans_hndl = (AEP_U64)getpid() << 32 | tran_id;

  /*This command is for the accelerator device*/
	message_hdr->flags = AEP_F_NOCTXCMD | AEP_F_SA;

	/*Set the message length field*/
	message_hdr->datalen = datalen;
	
	
	/*This transaction will always be synchronous*/
	rv = AEP_I_Transact(hConnection, &commandblock, &replyblock);
	
	if(rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GetVersion", rv, "AEP_GetVersion failed on connection %l", hConnection);
		/*Free the data block that was used to hold the message*/
		AEP_free(commandblock.ptr);
		return rv;
	}
	
	/*We make this assignment so we can refer to our message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)replyblock.ptr;
	
	/*Read back the datalen from the returned message*/
	datalen = message_hdr->datalen;
	
	/*replyIndex is used as an index to the reply so that we can parse it out*/
	replyIndex = replyblock.ptr + HDR_DATA_OFFSET;
		

	/*The target returns a number of strings that comprise it's version information.
	  Each string must be copied into the appropriate field in the version array*/
	/*Note: We use strncpy to ensure that we don't try to copy a string to an array 
	  that's too small for it. strncpy will append a null character to the end of the string if
	  there is room for it. But if, for example you call strncpy(dest, src, 20); where the src 
	  string is 20 characters long the null terminator is not appended. So we need to explicitly 
	  append a null character*/

	/*Target version*/
	strncpy(pVersions->target_version.target_version, (char *)replyIndex, sizeof(pVersions->target_version.target_version)-1);
	pVersions->target_version.target_version[sizeof(pVersions->target_version.target_version)-1]='\0';
	replyIndex+=strlen(replyIndex)+1;
	/*Target build date and time*/
	/*Note: the build date and time are sent as seperate strings, one after the other. We need to 
	  concatenate them to form the string for the API*/
	sprintf(pVersions->target_version.target_build_date, "%s %s", (char *)replyIndex, (char *)(replyIndex + strlen((char *)replyIndex)+1));
	pVersions->target_version.target_version[sizeof(pVersions->target_version.target_build_date)-1]='\0';
	replyIndex+=strlen(replyIndex)+1;
	replyIndex+=strlen(replyIndex)+1;
	/*Compiler version*/
	strncpy(pVersions->target_version.compiler_version, (char *)replyIndex, sizeof(pVersions->target_version.compiler_version)-1);
	pVersions->target_version.compiler_version[sizeof(pVersions->target_version.compiler_version)-1]='\0';
	replyIndex+=strlen(replyIndex)+1;
	/*Micro engine build date*/
	strncpy(pVersions->target_version.micro_engine_build_date, (char *)replyIndex, sizeof(pVersions->target_version.micro_engine_build_date)-1);
	pVersions->target_version.micro_engine_build_date[sizeof(pVersions->target_version.micro_engine_build_date)-1]='\0';
	replyIndex+=strlen(replyIndex)+1;
	/*CPLD version*/
	strncpy(pVersions->target_version.cpld_version, (char *)replyIndex, sizeof(pVersions->target_version.cpld_version)-1);
	pVersions->target_version.cpld_version[sizeof(pVersions->target_version.cpld_version)-1]='\0';
	replyIndex+=strlen(replyIndex)+1;
	/*Board Serial no. And Type*/
	strncpy(pVersions->target_version.board_serial_no_and_type, (char *)replyIndex, sizeof(pVersions->target_version.board_serial_no_and_type)-1);
	pVersions->target_version.board_serial_no_and_type[sizeof(pVersions->target_version.board_serial_no_and_type)-1]='\0';
	replyIndex+=strlen(replyIndex)+1;
	/*IXP1200 type*/
	strncpy(pVersions->target_version.ixp1200_type, (char *)replyIndex, sizeof(pVersions->target_version.ixp1200_type)-1);
	pVersions->target_version.ixp1200_type[sizeof(pVersions->target_version.ixp1200_type)-1]='\0';
	replyIndex+=strlen(replyIndex)+1;
	/*Thermal RNG version*/
	strncpy(pVersions->target_version.thermal_rng_version, (char *)replyIndex, sizeof(pVersions->target_version.thermal_rng_version)-1);
	pVersions->target_version.thermal_rng_version[sizeof(pVersions->target_version.thermal_rng_version)-1]='\0';
	replyIndex+=strlen(replyIndex)+1;
	/*Thermal RNG build date*/
	strncpy(pVersions->target_version.thermal_rng_build_date, (char *)replyIndex, sizeof(pVersions->target_version.thermal_rng_build_date)-1);
	pVersions->target_version.thermal_rng_build_date[sizeof(pVersions->target_version.thermal_rng_build_date)-1]='\0';
	replyIndex+=strlen(replyIndex)+1;
	/*Thermal RNG build time*/
	strncpy(pVersions->target_version.thermal_rng_build_time, (char *)replyIndex, sizeof(pVersions->target_version.thermal_rng_build_time)-1);
	pVersions->target_version.thermal_rng_build_time[sizeof(pVersions->target_version.thermal_rng_build_time)-1]='\0';
	/*replyIndex+=strlen(replyIndex);*/

	
	/*Free the data block that was used to hold the reply*/
	AEP_free(replyblock.ptr);

	/*Free the data block that was used to hold the message*/
	AEP_free(commandblock.ptr);

    	

	/*Step 2: Now, send a server command to the daemon to get the daemon version number*/
	
	/* Get the Tran ID to assign to this request*/
	/*Lock the flags Mutex: needed to protect the global transaction counter*/
	_LOCK(&AEPAPI_flags_mutex);
	rv = AEP_I_NewTranID(&tran_id);
	if(rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GetVersion", rv, "AEP_GetVersion failed on connection %l", hConnection);
		_UNLOCK(&AEPAPI_flags_mutex);
		return rv;
	}
	
	
	_UNLOCK(&AEPAPI_flags_mutex);
	
	datalen = 0; 

	/*How big is this byteblock going to be?*/
	commandblock.Len = HDR_SIZE + datalen;
	
	/*Allocate a buffer big enough to hold the message (Header + Data)*/
	commandblock.ptr=(AEP_U8_PTR)AEP_malloc(commandblock.Len);
	if(commandblock.ptr == NULL_PTR)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GetVersion", AEP_R_HOST_MEMORY, "Error: Host Memory Allocation Failure");
		return AEP_R_HOST_MEMORY;
	}

	/*We make this assignment so we can refer to our message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)commandblock.ptr;

	/*Copy the command identifier, the data length and the status*/
	message_hdr->command = AEP_C_GET_DAEMON_VERSION;
	
	message_hdr->status = AEP_S_INIT;

  	/*Copy the transaction handle for this request*/
	message_hdr->trans_hndl = (AEP_U64)getpid() << 32 | tran_id;

  	/*This command is for the daemon*/
	message_hdr->flags = AEP_F_SRVCMD;

	/*Set the message length field*/
	
	message_hdr->datalen = datalen;
	
	/*Call the internal function to get the daemon version*/
	rv = AEP_I_GetDaemonVersion(hConnection, &commandblock, &replyblock);
	if(rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_GetVersion", rv, "AEP_I_GetVersion failed to get Daemon Version Number %l", hConnection);
		/*Free the data block that was used to hold the message*/
		AEP_free(commandblock.ptr);
		return rv;
	}

	
	
	/*We make this assignment so we can refer to our message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)replyblock.ptr;
	
	datalen = message_hdr->datalen;
	
	/*Copy the reply to the right place and free up memory*/
	memcpy(&(pVersions->daemon_version), replyblock.ptr + HDR_DATA_OFFSET , sizeof(AEP_DAEMON_VERSION));
	AEP_free (commandblock.ptr);
	AEP_free (replyblock.ptr);


	/*Step 3: Insert the API version number*/
	/*Note: The API version number is defined in the file api_version.h*/
	/*Make sure that the API version string is not longer than the array we have to hold it (100 chars)*/

	strncpy(pVersions->api_version.versionString, API_VERSION, 99);
	strncat(pVersions->api_version.versionString, COMPILER_VERSION, 99 - strlen(API_VERSION));
	pVersions->api_version.versionString[99]='\0';
	AEP_TRACE(API, TRACE_DEBUG, FN_NONE, "AEP_GetVersion", AEP_R_OK, "...Leaving");
	
	
	return rv; 
}

 
