/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/


/****************************************************************/
/*								*/
/* Filename:	timeout_hndlr.c					*/
/* Description:	This thread traverses the response hash table 	*/
/*		at intervals and checks to see if any of the 	*/
/*		response items' time stamps are stale		*/
/*								*/
/* Copyright:	Accelerated Encryption Processing Ltd., 2000	*/
/* Author:	Diarmuid O'Neill				*/
/* Product:	AEP1000						*/
/* Date:	7/11/99						*/
/*								*/
/****************************************************************/

#include <unistd.h>
#include <pthread.h>

#include "aep_i_defs.h"
#include "aep_i_types.h"
#include "aep_daemon_defs.h"
#include "config.h"

#include "mutex.h"

#include "aepmalloc.h"

#include "sockets.h"
#include "hash.h"
#include "aeptrace.h"

#include "timeout_hndlr.h"

extern pthread_mutex_t response_item_mutex_array[];

/*This thread traverses the response hash table at intervals and checks to see if any
	of the response items' time stamps are stale*/

void *ptrTimedOutHndlr()
{
  /*Used to hold the current response item*/
	AEP_I_RESPONSE_DATA_PTR pResponse;
	
	AEP_I_HASH_BUCKET_PTR tmpBucket = NULL_PTR;

  /*Return value*/
	AEP_RV rv;

  /*Holds current time*/
	struct timeval timenow;
	
	int i;

  /*Check the hash table for ever*/	
	while(1)
	{
		/*Get current time*/
		gettimeofday(&timenow, NULL);
	
		for (i=0; i<ResponseTableSize; i++)
		{
/*Lock the response table*/
			_LOCK(&response_item_mutex_array[i]);
			
		  /*Get a pointer to the first entry in this row (if any)*/
			tmpBucket = AEP_I_response_table.response_table->table[i];			
		
		  /*While there's more entries in this row*/	
			while(tmpBucket != NULL_PTR)
			{
			  /*Get the response item pointer*/
				pResponse = tmpBucket->val;
			
			  /*Check it's timestamp against the current time*/
				if ( (timenow.tv_sec > pResponse->timestamp.tv_sec) || 
						( (timenow.tv_sec == pResponse->timestamp.tv_sec) && 
						(timenow.tv_usec > pResponse->timestamp.tv_usec) ) )
				{
					/*If there's a valid semaphore (ie someones waiting on this transaction)
					then signal them to let them know the answer has arrived*/
					if(pResponse->response_semaphore != NULL)
					{	
					  /*Make the response item as timed out*/
						((AEP_MSG_HDR_PTR)pResponse->msg)->status = AEP_S_TIMEDOUT;
						
						AEP_I_SendToSocket(pResponse->sock_fd, (AEP_U8*)pResponse->msg,
							((AEP_MSG_HDR_PTR)pResponse->msg)->datalen + HDR_SIZE);

					  /*Move tmpBucket on to the next entry*/
						tmpBucket = tmpBucket->next;
							
					  /*Remove the response entry from the hash table*/
						AEP_I_HashRemoveEntry(AEP_I_response_table.response_table,pResponse->transaction_handle);
				  
					  /*Signal the semaphore associated with this response*/
						sem_post(pResponse->response_semaphore);

						AEP_TRACE (DAEMON, TRACE_WARNINGS, FN_NONE, "ptrTimedOutHndlr", AEP_R_TIMED_OUT, "Response Item Timed Out, Semaphore Signalled");
					}
					else  /*No one is waiting on this response*/
					{
					  /*Move tmpBucket on to the next entry*/
						tmpBucket = tmpBucket->next;
					  /*Remove the entry from the responses table*/
						rv = AEP_I_HashRemoveEntry(AEP_I_response_table.response_table,pResponse->transaction_handle);
						if(rv != AEP_R_OK)
							AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "ptrTimedOutHndlr", rv, "Error removing Response from Hash Table");
										
					  /*Free the mem used to hold the msg part of the response*/
						AEP_free(pResponse->msg);

					  /*Free the mem used to hold the response*/
						AEP_free (pResponse);

						AEP_TRACE (DAEMON, TRACE_WARNINGS, FN_NONE, "ptrTimedOutHndlr", AEP_R_TIMED_OUT, "Response Item Timed Out");
					}
				}
				else		
			  	 /*Move on to the next entry in this list*/
					tmpBucket = tmpBucket->next;
		
			}

/*UnLock the response table*/
			_UNLOCK(&response_item_mutex_array[i]);
	
		}

	  /*Wait a while before we check again*/
		sleep(5);
	}
}
